@@ -3,6 +3,8 @@ from itertools import combinations
 from pyscf import gto, scf, symm, tools, ao2mo
 from openfermion import QubitOperator, FermionOperator, jordan_wigner, utils, linalg
 from qiskit import opflow, quantum_info
+from qiskit_nature.operators.second_quantization import SecondQuantizedOp, FermionicOp
+from qiskit_nature.mappers.second_quantization import JordanWignerMapper
 from IPython.display import display, HTML
 from tabulate import tabulate
 
@@ -220,7 +222,7 @@ def find_symmetry_generators(mol, mf, irrep):
     beta_labels = str(beta_labels).replace('\'', '')
     alpha_labels = str(alpha_labels).replace('\'', '')
 
-    #signs of Pα and Pβ
+    #signs of P↑ and P↓
     number_up = (mol.nelectron + mol.spin)//2
     number_down = (mol.nelectron - mol.spin)//2
     
@@ -237,11 +239,11 @@ def find_symmetry_generators(mol, mf, irrep):
     symmetry_generators.insert(0, n_alpha)
     symmetry_generators.insert(1, n_beta)
 
-    symmetry_generator_labels.insert(0, 'Pα')
+    symmetry_generator_labels.insert(0, 'P↑')
     symmetry_generator_labels.insert(1, 'P\'')
 
-    descriptions.insert(0, 'Parity of electrons with spin α')
-    descriptions.insert(1, 'Parity of electrons with spin α for orbitals in irrep ' + alpha_labels[1:-1] + ' and with spin β for orbitals in irrep ' + beta_labels[1:-1])
+    descriptions.insert(0, 'Parity of electrons with spin up')
+    descriptions.insert(1, 'Parity of electrons with spin up for orbitals in irrep ' + alpha_labels[1:-1] + ' and with spin down for orbitals in irrep ' + beta_labels[1:-1])
 
     irreps_string = str()
     rep_numbers = [0]*len(irrep_labels)
@@ -426,6 +428,39 @@ def eliminate_qubits(qubit_operator, target_qubits):
             output.terms[tuple(new_key)] = qubit_operator.terms[keys]
     return output
 
+def project_operator(qubit_operator, target_qubits):
+    """Keeps only the terms that commute with the symmetries from a qubit operator, and eliminates the target qubits from them
+
+    Args:
+        qubit_operator (openfermion.QubitOperator): operator to eliminate qubits from
+        target_qubits (list): list of integers that correspond to the target qubits to eliminate
+
+    Returns:
+        openfermion.QubitOperator: the operator without the target qubits
+    """
+    output = QubitOperator()
+    #creates a list of qubits that will not be eliminated
+    new_qubits = list()
+    for qubit in range(utils.count_qubits(qubit_operator)):
+        if qubit not in target_qubits:
+           new_qubits.append(qubit)
+    #eliminates the target qubits from the input qubit operator, and returns 
+    for keys in list(qubit_operator.terms.keys()):
+        new_key = list()
+        count = 0
+        for key in keys:
+            if key[0] not in target_qubits:
+                new_key.append((new_qubits.index(key[0]), key[1]))
+            elif key[1] == 'X' or key[1] == 'Y':
+                count += 1
+                break
+        if count == 0:
+            try:
+                output.terms[tuple(new_key)] += qubit_operator.terms[keys]
+            except:
+                output.terms[tuple(new_key)] = qubit_operator.terms[keys]
+    return output
+
 def get_hamiltonian(mol, mf):
     """Constructs the second-quantized molecular Hamiltonian
 
@@ -484,6 +519,7 @@ def get_MO_irreps_string(label_orb_symm, irrep_labels, html = False):
 
     Returns:
         str: the molecular orbital irreps in the format: '4b1u 2b3g 3b1u 4ag 2b2u 1b2g 1b3u 1b3g 3ag 1b2u 2b1u 2ag 1b1u 1ag'
+        array: the 1D array that contains the molecular orbital names labelled by their irrep
     """
     MO_irreps_string = str()
     orbital_names = list()
@@ -697,6 +733,18 @@ def QubitOperator_to_PauliSumOp(qubitoperator):
         output += qubitoperator.terms[key]*quantum_info.SparsePauliOp(string)
     return opflow.PauliSumOp(output).reduce()
 
+def PauliSumOp_to_QubitOperator():
+    output = QubitOperator()
+    for k in x:
+        k = str(k.to_pauli_op())
+        N = len(k)
+        term = str()
+        for i, s in enumerate(k):
+            if s == 'X' or s == 'Z' or s == 'Y':
+                term += s + str(N - i) + ' '
+        output += QubitOperator(term)
+    return output
+
 def find_ground_state_irrep(orbital_labels, mo_occ, character_table, irrep_labels):
 
     """Finds the ground state irrep
@@ -719,13 +767,14 @@ def find_ground_state_irrep(orbital_labels, mo_occ, character_table, irrep_label
             irrep = irrep_labels[i]
     return irrep
 
-def qubit_table(symmetry_generators, signs, target_qubits):
-    """Creates an HTML table showing how the symmetry-adapted encoding stores information about occupancy in each qubit
+def qubit_table(symmetry_generators, signs, target_qubits, orbital_names):
+    """Creates an HTML table showing how the symmetry-adapted encoding stores information about occupancy of the spin-orbitals in each qubit
 
     Args:
         symmetry_generators (array): the input symmetry generators encoded as arrays of 1s (identiy) and -1 (Z Pauli operator)
         signs (array): the signs of the corresponding generators in symmetry_generators encoded as an array of 1s (+) and -1s (-)
         target_qubits (array): the target qubits of the corresponding generators to eliminate
+        orbital_names (array): the 1D array that contains the molecular orbital names labelled by their irrep
 
     Returns:
         str: the table in HTML format
@@ -735,7 +784,10 @@ def qubit_table(symmetry_generators, signs, target_qubits):
     orbitals_column = list()
     qubit_column = list()
     for i in range(n):
-        orbitals_column.append(i)
+        if i % 2 == 0:
+            orbitals_column.append(orbital_names[i//2][:-7] + '↑' + orbital_names[i//2][-7:])
+        else:
+            orbitals_column.append(orbital_names[i//2][:-7] + '↓' + orbital_names[i//2][-7:])
         if i in target_qubits:
             string = str()
             for j in range(n):
@@ -755,7 +807,11 @@ def qubit_table(symmetry_generators, signs, target_qubits):
                 if t < i:
                     nt += 1
             qubit_column.append('q' + str(i - nt).translate(subscripts))
-    return tabulate(np.array([orbitals_column, qubit_column]).T, headers = ['Spin-orbital', 'Occupancy from qubits'], tablefmt='html')
+    output = '<table><tr><td><b>Spin-orbital</b></td><td><b>Occupancy from qubits</b></td></tr>'
+    for i in range(n):
+        output += '<tr><td>' + orbitals_column[i] + '</td><td>' + qubit_column[i] + '</td></tr>'
+    output += '</table>'
+    return output
 
 def make_encoding(atom, basis, charge = 0, spin = 0, irrep = None):
     """Makes an array that stores information about that encoding and which can be passed on to other functions
@@ -806,9 +862,15 @@ def apply_encoding(operator, encoding):
     """
     if type(operator) == FermionOperator:
         operator = jordan_wigner(operator)
+    if type(operator) == SecondQuantizedOp:
+        operator = JordanWignerMapper.map(operator)
+    if type(operator) == FermionicOp:
+        operator = JordanWignerMapper.map(operator)
+    if type(operator) == opflow.PauliSumOp:
+        operator = PauliSumOp_to_QubitOperator(operator)
     tableau, tableau_signs, target_qubits = encoding
     transformed_operator = apply_Clifford_tableau(operator, tableau, tableau_signs)
-    reduced_operator = simplify_QubitOperator(eliminate_qubits(transformed_operator, target_qubits))
+    reduced_operator = simplify_QubitOperator(project_operator(transformed_operator, target_qubits))
     return reduced_operator
 
 def reduced_hamiltonian(atom, basis, charge = 0, spin = 0, irrep = None, verbose = True, show_lowest_eigenvalue = False, output_format = 'openfermion'):
@@ -891,23 +953,23 @@ def reduced_hamiltonian(atom, basis, charge = 0, spin = 0, irrep = None, verbose
         '<h2>Molecular orbitals</h2>' +
         '<p>The molecular orbitals (MOs) for ' + molecule_name + ' in the ' + mol.basis + ' basis are:</p>' +
         MO_irreps_string +
-        '<p>Each molecular orbital corresponds to two spin-orbitals and to two qubits in the Jordan-Wigner basis (in the order β, α), for a total of ' + str(len(label_orb_symm)) + ' orbitals and ' + str(qubits1) + ' qubits before qubit number reduction.</p>' +
+        '<p>Each molecular orbital corresponds to two spin-orbitals and to two qubits in the Jordan-Wigner basis (in the order ↓, ↑), for a total of ' + str(len(label_orb_symm)) + ' orbitals and ' + str(qubits1) + ' qubits before qubit number reduction.</p>' +
         '<h2>Full Boolean symmetry group</h1>' +
-        '<p>The operators for the number of electrons in spin α (N<sub>α</sub>) and β (N<sub>β</sub>) also correspond to symmetries of the Hamiltonian,' + 
-        'and therefore so do the parity operators P<sub>α</sub> = (-1)<sup>N<sub>α</sub></sup> and P<sub>β</sub> = (-1)<sup>N<sub>β</sub></sup> (which generate a symmetry group isomorphic to ℤ<sub>2</sub><sup>2</sup>).</p>' +
+        '<p>The operators for the number of electrons in spin up (N<sub>↑</sub>) and down (N<sub>↓</sub>) also correspond to symmetries of the Hamiltonian,' + 
+        'and therefore so do the parity operators P<sub>↑</sub> = (-1)<sup>N<sub>↑</sub></sup> and P<sub>↓</sub> = (-1)<sup>N<sub>↓</sub></sup> (which generate a symmetry group isomorphic to ℤ<sub>2</sub><sup>2</sup>).</p>' +
         '<p>The full Boolean symmetry group is isomorphic to ℤ <sub>2</sub><sup>' + str(len(target_qubits)) + '</sup> and its ' + str(len(target_qubits)) + ' generators and their representations in the qubit space are then:</p>' +
         tabulate(zip(symmetry_generator_labels, descriptions, symmetry_generators_strings, eigensectors, target_qubits), headers = ['Symmetry operator', 'Description', 'Qubit representation', 'Eigensector', 'Qubit to remove'], tablefmt='html') +
-        '<p>The target eigensectors are the ones of states that are in the ' + irrep[0] + '<sub>' + irrep[1:] + '</sub> irreducible representation of ' + html_groupname + ', and for which N<sub>α</sub> is ' + str((mol.nelectron + mol.spin)//2 % 2).replace('0', 'even').replace('1', 'odd') + ' and N<sub>β</sub> is ' + str((mol.nelectron - mol.spin)//2 % 2).replace('0', 'even').replace('1', 'odd') + '.</p>' +
+        '<p>The target eigensectors are the ones of states that are in the ' + irrep[0] + '<sub>' + irrep[1:] + '</sub> irreducible representation of ' + html_groupname + ', and for which N<sub>↑</sub> is ' + str((mol.nelectron + mol.spin)//2 % 2).replace('0', 'even').replace('1', 'odd') + ' and N<sub>↓</sub> is ' + str((mol.nelectron - mol.spin)//2 % 2).replace('0', 'even').replace('1', 'odd') + '.</p>' +
         '<h2>Change-of-basis transformation and Clifford tableau</h2>' +
         '<p>The change-of-basis transformation acts on the qubit operators corresponding to the generators of the Boolean symmetry group as:</p>' +
         clifford_on_symmetries +
         '<p>Its Clifford tableau is:</p>' +
         show_tableau(tableau, tableau_signs, html=True) +
         show_Clifford_kets_HTML(tableau, tableau_signs, mf.mo_occ, target_qubits, full = full) +
-        '<p>The states in the target eignesectors all have qubits ' + str(target_qubits)[1:-1] + ' equal to |0⟩, and these qubits are removed to obtained the <b>symmetry-adapted encoding</b>.</p>' +
+        '<p>The states in the target eignesectors all have qubits ' + str(target_qubits)[1:-1] + ' equal to |0⟩, and these qubits are removed to obtain the <b>symmetry-adapted encoding</b>.</p>' +
         '<h2>Qubits in the symmetry-adapted encoding</h2>' +
         '<p>In the <b>symmetry-adapted encoding</b> information about the occupancy of the ' + str(qubits1) + ' spin-orbitals is encoded in ' + str(qubits3) + ' qubits as:</p>' +
-        qubit_table(target_qubits, signs, symmetry_generators) +
+        qubit_table(symmetry_generators, signs, target_qubits, orbital_names) +
         '<p><i>(where addition is taken modulo 2)</i></p>' +
         '<h2>Jordan-Wigner Hamiltonian</h2>' +
         '<p>The <b>second-quantised qubit Hamiltonian</b> in the <b>Jordan-Wigner basis</b> is given by:</p>' +
@@ -929,6 +991,6 @@ def reduced_hamiltonian(atom, basis, charge = 0, spin = 0, irrep = None, verbose
         return reduced_hamiltonian
     elif output_format == 'qiskit':
         reduced_hamiltonian = QubitOperator_to_PauliSumOp(reduced_hamiltonian)
-        print(reduced_hamiltonian)
+        return reduced_hamiltonian
     else:
         raise ValueError('The output format must be either OpenFermion or Qiskit')
\ No newline at end of file
